import datetime
import hashlib
import itertools
import math
import string
import struct
import uuid

try:
    from typing_extensions import TypeAlias
except ImportError:
    from typing import TypeAlias

from typing import Annotated, Dict, List, Union

from johen.examples import Examples
from johen.random import gen

simple_symbol = ("".join(r.sample(string.ascii_letters + "_", 5)) for r in gen)
unsigned_ints = (r.getrandbits(2 ** r.randint(0, 6)) for r in gen)
negative_ints = (i * -1 for i in unsigned_ints)
ints = gen.one_of(unsigned_ints, negative_ints, (0,))
bit_packed_floats = (struct.unpack("d", struct.pack("Q", bits))[0] for bits in unsigned_ints)
valid_floats = (v for v in bit_packed_floats if not math.isinf(v) and not math.isnan(v))
invalid_floats = gen.one_of((math.inf, math.nan))
all_floats = gen.one_of(valid_floats, invalid_floats)
bools = gen.one_of([True, False])
objects = (object() for _ in gen)
printable_strings = ("".join(r.sample(string.printable, r.randint(0, 6))) for r in gen)
colors = gen.one_of(
    [
        "red",
        "green",
        "blue",
        "orange",
        "purple",
        "cyan",
        "magenta",
        "magenta",
        "yellow",
        "gold",
        "silver",
        "black",
        "white",
    ]
)
things = gen.one_of(
    [
        "shirt",
        "sneaker",
        "shoe",
        "apple",
        "banana",
        "orange",
        "tea",
        "sandwich",
        "tennis",
        "football",
        "basketball",
        "fork",
        "table",
        "computer",
    ]
)
names = gen.one_of(
    [
        "bob",
        "alice",
        "jennifer",
        "john",
        "mary",
        "jane",
        "sally",
        "fred",
        "dan",
        "alex",
        "margaret",
        "vincent",
        "timothy",
        "samuel",
    ]
)
ascii_words = ("-".join(group) for group in zip(colors, names, things))
dates = (
    datetime.date(2013, 1, 1)
    + datetime.timedelta(
        days=r.randint(0, 365 * 20),
    )
    for r in gen
)
datetimes = (
    datetime.datetime(2013, 1, 1, 1)
    + datetime.timedelta(
        days=r.randint(0, 365 * 20),
        seconds=r.randint(0, 60 * 60 * 24),
        milliseconds=r.randint(0, 1000),
    )
    for r in gen
)
positive_timedeltas = (
    datetime.timedelta(seconds=r.randint(0, 59), hours=r.randint(0, 23)) for r in gen
)
uuids = (uuid.UUID(int=r.getrandbits(128), version=4) for r in gen)
uuid_hexes = (uid.hex for uid in uuids)
file_extensions = gen.one_of(
    (".jpg", ".png", ".gif", ".txt", ".py", ".ts", ".c", ".obj", ".ini", "")
)
path_segments = gen.one_of(
    (
        ".",
        "..",
        "tmp",
        "var",
        "usr",
        "Home",
        "data",
        "volumes",
        "etc",
        "tests",
        "src",
        "db",
        "conf",
        "events",
        "utils",
        "app",
        "versions",
        "models",
    ),
    uuid_hexes,
)
file_paths = (
    lead + "/".join(segment for _, segment in segments) + ext
    for lead, segments, ext in zip(
        gen.one_of(("", "/")),
        (zip(range(r.randint(1, 8)), path_segments) for r in gen),
        file_extensions,
    )
)
file_names = (segment + ext for segment, ext in zip(path_segments, file_extensions))
byte_strings = (s.encode("utf8") for s in printable_strings)
sha1s = (hashlib.sha1(s).hexdigest() for s in byte_strings)
json_primitives = gen.one_of(ascii_words, ints, bools, valid_floats, itertools.repeat(None))
nones = itertools.repeat(None)


UnsignedInt = Annotated[int, Examples(unsigned_ints)]
NegativeInt = Annotated[int, Examples(negative_ints)]
ValidFloat = Annotated[float, Examples(valid_floats)]
InvalidFloat = Annotated[float, Examples(invalid_floats)]
ValidOrInvalidFloat = Annotated[bool, Examples(all_floats)]
AsciiWord = Annotated[str, Examples(ascii_words)]
FileName = Annotated[str, Examples(file_names)]
FilePath = Annotated[str, Examples(file_paths)]
Sha1 = Annotated[str, Examples(sha1s)]
SimpleSymbol = Annotated[str, Examples(simple_symbol)]


JsonValue: TypeAlias = Union[int, float, str, bool, None, List["JsonValue"], "JsonDict"]
JsonDict: TypeAlias = Dict[str, JsonValue]



import dataclasses
import functools
import inspect
import typing
from collections import defaultdict
from typing import Any

import pytest

from johen.config import ParametrizeConfig, compile_matchers, pick_seed_from_name, updated_config
from johen.exc import GenerationError
from johen.generators.annotations import AnnotationProcessingContext
from johen.generators.base import generate_dicts_for_annotations
from johen.globals import global_config
from johen.random import gen

_C = typing.TypeVar("_C", bound=typing.Callable)

try:
    from typing_extensions import Unpack
except ImportError:
    from typing import Unpack


@dataclasses.dataclass
class _parametrize:
    config: list[ParametrizeConfig] = dataclasses.field(default_factory=lambda: [global_config])

    @typing.overload
    def __call__(self, test: _C, **kwargs: Unpack[ParametrizeConfig]) -> "_C":
        pass

    @typing.overload
    def __call__(self, test: None = None, **kwargs: Unpack[ParametrizeConfig]) -> "_parametrize":
        pass

    def __call__(
        self, test: _C | None = None, **kwargs: Unpack[ParametrizeConfig]
    ) -> "_parametrize | _C":
        if kwargs:
            p = _parametrize([*self.config, kwargs])
        else:
            p = self

        if test is None:
            return p

        configs = p.config
        gather_config: ParametrizeConfig = functools.reduce(updated_config, configs)
        argspec = inspect.getfullargspec(test)
        injected_args = sorted(
            gather_config["arg_set"] if gather_config["arg_set"] is not None else argspec.args
        )

        if not injected_args:
            return test

        if gather_config["seed"] is None:
            final_seed = pick_seed_from_name(test.__name__)
        else:
            final_seed = gather_config["seed"]

        count = gather_config["count"]
        assert count > 0, "count must be greater than 0"

        @functools.cache
        def get_call_args() -> typing.Iterator[dict]:
            final_config: ParametrizeConfig = functools.reduce(
                updated_config, [global_config, *configs[1:]]
            )

            if invalid_arg := next(
                (k not in injected_args for k in final_config["overrides"].keys()), None
            ):
                raise ValueError(
                    f"Argument {invalid_arg} cannot be overriden, check your arg_set and overrides arguments to parametrize."
                )

            context = AnnotationProcessingContext.from_source(test)
            context.generate_defaults = final_config["generate_defaults"]
            context.matchers = compile_matchers(final_config)
            context.globals = final_config["globals"]
            return gen.wrap_deterministically(
                generate_dicts_for_annotations(
                    {
                        k: final_config["overrides"].get(k, argspec.annotations.get(k, Any))
                        for k in injected_args
                    },
                    context,
                    optional_keys=[],
                ),
                seed=final_seed,
                max_iterations=final_config["max_iterations"],
            )

        cached: list[dict[str, Any]] = []
        seeds: list[int] = []

        def _get_arg_slice(arg: str) -> typing.Callable[[int], typing.Callable[[], Any]]:
            def _get_arg_thunk(index: int):
                def _thunk():
                    call_args = get_call_args()
                    while len(cached) < count:
                        try:
                            cached.append(next(call_args))
                            seeds.append(gen.last_seed)
                        except StopIteration as e:
                            raise GenerationError(
                                f"Failed to generate {count} test cases for {test.__name__}, check that constraint is not too strong."
                            ) from e
                    gen.restart_at(seeds[index])
                    return cached[index][arg]

                _thunk.__name__ = f"{arg}-{index}"
                return _thunk

            return _get_arg_thunk

        return pytest.mark.johen(count, injected_args)(
            pytest.mark.parametrize(
                injected_args,
                [
                    [_get_arg_slice(k)(i) for k in injected_args]
                    for i in range(gather_config["count"])
                ],
                indirect=(),
            )(test)
        )


def pytest_configure(config: pytest.Config):
    config.addinivalue_line(
        "markers", "johen(injected): marks a test for parametrization via the johen.pytest hooks."
    )


_default = object()


@dataclasses.dataclass
class Sometimes:
    calls: dict[tuple[str, int], tuple[set[int], set[int]]] = dataclasses.field(
        default_factory=lambda: defaultdict(lambda: (set(), set()))
    )

    def __call__(self, cond: Any) -> bool:
        result = bool(cond)
        frameinfo = inspect.getouterframes(inspect.currentframe())[1]

        hits, misses = self.calls[(frameinfo.filename, frameinfo.lineno)]

        if not result:
            misses.add(gen.last_seed)
            return True

        hits.add(gen.last_seed)
        return True


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_setup(item: pytest.Item):
    if not isinstance(item, pytest.Function):
        yield
        return

    mark = item.get_closest_marker("johen")
    if mark is None:
        yield
        return

    injected: list[str] = mark.args[1]

    if hasattr(item, "callspec"):
        for k in injected:
            # apply the thunk.
            item.callspec.params[k] = item.callspec.params[k]()  # type: ignore
    else:
        raise GenerationError(
            f"Test {item.name!r} does not support parametrization, you may need to use a plain test function."
        )

    yield


sometimes = Sometimes()


_runs_remaining = 0
_last_item = -1


@pytest.fixture(autouse=True)
def check_sometimes_misses(request: pytest.FixtureRequest):
    global _last_item, _runs_remaining

    node = request.node
    marker = node.get_closest_marker("johen")
    if not isinstance(node, pytest.Function) or not marker:
        yield
        return

    fid = id(node.obj)

    if fid != _last_item:
        _last_item = fid
        _runs_remaining = marker.args[0]
        sometimes.calls.clear()

    yield

    _runs_remaining -= 1
    if _runs_remaining > 0:
        return

    failure_lines = []
    for (filename, lineno), (hits, misses) in sometimes.calls.items():
        if len(hits) == marker.args[0]:
            failure_lines.append(
                f"{filename}:{lineno} -- all tests hit, try increasing count to find counterfactuals"
            )
        elif not hits:
            failure_lines.append(f"{filename}:{lineno} -- no hits, try different seed or count")
    if failure_lines:
        assert False, "sometimes failures\n" + "\n".join(failure_lines)


parametrize = _parametrize()



from typing import Any, Iterator, get_type_hints

from pydantic import BaseModel
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined

from johen.generators.annotations import AnnotationProcessingContext
from johen.generators.base import generate_dicts_for_annotations


def generate_dicts_for_pydantic_model(
    context: "AnnotationProcessingContext",
) -> Iterator[dict[str, Any]]:
    hints = get_type_hints(context.source, include_extras=True)
    return generate_dicts_for_annotations(
        {k: hints.get(k, Any) for k, field in context.source.model_fields.items()},
        context,
        optional_keys=[
            k for k, field in context.source.model_fields.items() if _pydantic_has_default(field)
        ],
    )


def generate_pydantic_instances(context: AnnotationProcessingContext) -> Iterator[BaseModel] | None:
    if isinstance(context.source, type) and issubclass(context.source, BaseModel):
        return (context.source(**d) for d in generate_dicts_for_pydantic_model(context))
    return None


def _pydantic_has_default(field: FieldInfo) -> bool:
    return field.default is not PydanticUndefined or field.default_factory is not None



import dataclasses
import inspect
import typing
from typing import Any, Iterator

from johen.exc import GenerationError
from johen.random import gen

_A = typing.TypeVar("_A")


class AnnotationMatcher(typing.Protocol):
    def __call__(self, context: "AnnotationProcessingContext") -> Iterator[Any] | None:
        ...


@dataclasses.dataclass
class AnnotationProcessingContext:
    source: Any
    origin: Any | None
    args: tuple[Any, ...]
    path: tuple[str, ...] = dataclasses.field(default_factory=tuple)
    generate_defaults: bool | typing.Literal["holes"] = False
    matchers: list[AnnotationMatcher] = dataclasses.field(default_factory=list)
    globals: dict[str, Any] = dataclasses.field(default_factory=dict)
    recursive_depth: int = 0

    def concretely_implements(self, other: Any) -> bool:
        for origin in (self.origin, self.source):
            if origin is None or not inspect.isclass(origin):
                continue
            return issubclass(origin, other)
        return False

    def wrap_with_debug_context(self, iterator: Iterator[_A]) -> Iterator[_A]:
        def wrapper():
            try:
                for i in iterator:
                    yield i
            except Exception as e:
                raise GenerationError(
                    f"Generation failed unexpectedly for {' '.join(self.path)} {self.source}"
                ) from e

        return wrapper()

    def concretely_implemented_by(self, other: Any) -> bool:
        for origin in (self.origin, self.source):
            if origin is None:
                continue
            try:
                return issubclass(other, origin)
            except TypeError:
                continue
        return False

    def step(
        self,
        source: Any,
        step: str | None = None,
        args: tuple[Any, ...] | None = None,
        recursive=False,
    ) -> Iterator:
        next_context = AnnotationProcessingContext.from_source(source)
        if args is not None:
            next_context.origin = source
            next_context.args = args
        next_context.path = (*self.path, step) if step else self.path
        next_context.generate_defaults = self.generate_defaults
        next_context.matchers = self.matchers
        next_context.globals = self.globals
        if recursive:
            next_context.recursive_depth = self.recursive_depth + 1

        return next_context.generate()

    @classmethod
    def from_source(cls, source: Any) -> "AnnotationProcessingContext":
        return AnnotationProcessingContext(
            source=source,
            origin=typing.get_origin(source),
            args=typing.get_args(source) or (),
            path=(repr(source),),
        )

    def generate(
        self,
    ) -> typing.Iterator:
        for matcher in self.matchers:
            result = matcher(self)
            if result is not None:
                return result
        raise GenerationError(f"Could not generate for {' '.join(self.path)} {self.source}")



import dataclasses


@dataclasses.dataclass
class NamedBool:
    message: str
    result: bool

    def __bool__(self):
        return self.result

    def __str__(self):
        return self.message

    __repr__ = __str__



import datetime
import random
import uuid
import zlib
from typing import Any, Iterable, Iterator, Literal, Type

from johen.generators import base, specialized
from johen.generators.annotations import AnnotationMatcher
from johen.random import gen

__all__ = [
    "ParametrizeConfig",
    "get_base_config",
    "updated_config",
    "compile_matchers",
    "pick_seed_from_name",
]

try:
    from typing_extensions import TypedDict
except ImportError:
    from typing import TypedDict


class ParametrizeConfig(TypedDict, total=False):
    """
    Configures how the parametrize family of functions behave.
    """

    # Controls how the seed should be set before generating examples.  When None, a default strategy is taken based on
    # parameterize implementation, usually hashing the test function name.  This should be stable so that local and CI
    # test runs agree.
    seed: int | None
    # Configures the number of parametrized examples that should be generated.
    count: int
    # See `AnnotationProcessingContext.generate_defaults`
    generate_defaults: bool | Literal["holes"]
    # Which named arguments to actually parametrize -- useful for excluding arguments that are provided by the testing
    # framework, such as fixtures or mocks.
    arg_set: Iterable[str] | None
    # Overrides for named parameters, a shorthand for using Annotated[x, Examples(iterator)]
    overrides: dict[str, Iterable]

    # The maximum calls to `gen.__next__` allowed before yielding a value.  Note that thanks to the halting
    # problem, there is no perfect answer here -- it's at the intersection of every way *you* use generators.
    # Ideally there is a maximum depth at which you are generating, and you find that through experience.
    # Increasing this can hide bugs and performance issues.  Decreasing this can lead to fragile generation and
    # unexpected failures.
    max_iterations: int
    type_matchers: dict[Any, Iterator]
    matchers: list[AnnotationMatcher]
    globals: dict[str, Any]


def pick_seed_from_name(name: str) -> int:
    return zlib.crc32(name.encode("utf8")) & 0xFFFFFFFF


def compile_matchers(config: ParametrizeConfig) -> list[AnnotationMatcher]:
    return [
        *config["matchers"],
        lambda context: config["type_matchers"].get(context.source, None),
    ]


def get_base_config() -> ParametrizeConfig:
    return {
        "seed": None,
        "generate_defaults": False,
        "arg_set": None,
        "overrides": {},
        "max_iterations": 10000,
        "count": 10,
        "type_matchers": {
            int: specialized.ints,
            str: specialized.ascii_words,
            uuid.UUID: specialized.uuids,
            bool: specialized.bools,
            object: specialized.objects,
            float: specialized.valid_floats,
            bytes: specialized.byte_strings,
            datetime.date: specialized.dates,
            datetime.datetime: specialized.datetimes,
            datetime.timedelta: specialized.positive_timedeltas,
            None: specialized.nones,
            type(None): specialized.nones,
            Any: specialized.json_primitives,
            random.Random: (r for r in gen),
        },
        "matchers": [
            base.generate_dicts_from_typeddict,
            base.generate_lists_sets_frozen_sets,
            base.generate_named_tuples,
            base.generate_dataclass_instances,
            base.generate_forward_refs,
            base.generate_literals,
            base.generate_dicts,
            base.generate_unions,
            base.generate_annotated,
            base.generate_enums,
            base.generate_tuples,
            base.generate_unexpected_annotation,
        ],
        "globals": {},
    }


def updated_config(left: ParametrizeConfig, right: ParametrizeConfig) -> ParametrizeConfig:
    default = get_base_config()

    return {
        "seed": right.get("seed", left.get("seed")),
        "count": right.get("count", left.get("count", default["count"])),
        "generate_defaults": right.get(
            "generate_defaults", left.get("generate_defaults", default["generate_defaults"])
        ),
        "arg_set": right.get("arg_set", left.get("arg_set")),
        "overrides": right.get("overrides", left.get("overrides", default["overrides"])),
        "max_iterations": right.get(
            "max_iterations", left.get("max_iterations", default["max_iterations"])
        ),
        "type_matchers": {**left.get("type_matchers", {}), **right.get("type_matchers", {})},
        "matchers": [*right.get("matchers", []), *left.get("matchers", [])],
        "globals": {**left.get("globals", {}), **right.get("globals", {})},
    }



from typing import Any, Iterable


class Examples:
    """
    Annotate types with this item in order to get example inputs as part of test generation.
    """

    def __init__(self, *args: Iterable[Any]):
        self.args = args

    def __iter__(self):
        return iter(self.args)

    def __str__(self):
        return f"Examples({self.args!r})"

    __repr__ = __str__



import inspect
import typing
from typing import Any, get_type_hints

import sqlalchemy.orm

from johen.generators.annotations import AnnotationProcessingContext
from johen.generators.base import generate_dicts_for_annotations


def generate_sqlalchemy_instance(
    context: AnnotationProcessingContext,
) -> typing.Iterator[Any] | None:
    import sqlalchemy.orm

    if inspect.isclass(context.source) and issubclass(
        context.source, sqlalchemy.orm.DeclarativeBase
    ):
        hints = get_type_hints(context.source, include_extras=True)
        inspection = sqlalchemy.inspect(context.source)
        dict_generator = generate_dicts_for_annotations(
            {
                c.key: next(iter(typing.get_args(hint)), Any)  # type: ignore
                if typing.get_origin(hint) is sqlalchemy.orm.Mapped
                else hint
                for c in inspection.c
                for hint in (hints.get(c.key, Any),)
            },
            context,
            optional_keys=[
                c.key
                for c in inspection.c
                if (
                    c.primary_key
                    or c.nullable
                    or c.default is not None
                    or c.server_default is not None
                )
            ],
        )

        return (context.source(**d) for d in dict_generator)
    return None



import dataclasses
import typing
from typing import Any

from johen.named_bool import NamedBool


class _Unset:
    pass


_unset = _Unset()


@dataclasses.dataclass
class change_watcher:
    cb: typing.Callable[[], Any]
    stack: list[Any] = dataclasses.field(default_factory=list)

    def __enter__(self):
        self.stack.append(ChangeResult(orig=self.cb()))
        return self.stack[-1]

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            return None

        self.stack.pop().result = self.cb()


@dataclasses.dataclass
class ChangeResult:
    orig: Any = _unset
    result: Any = _unset

    def from_value(self, value: Any):
        return NamedBool(f"{self} (expected from: {value!r})", bool(self and self.orig == value))

    def to_value(self, value: Any):
        return NamedBool(
            f"{self} (expected result: {value!r})", bool(self and self.result == value)
        )

    def __bool__(self) -> bool:
        assert (
            self.orig is not _unset
        ), "ChangeWatcher.__enter__ was not called, cannot compute result!"
        assert (
            self.result is not _unset
        ), "ChangeWatcher.__exit__ was not called, cannot compute result!"
        return self.result != self.orig

    def __str__(self) -> str:
        assert (
            self.orig is not _unset
        ), "ChangeWatcher.__enter__ was not called, cannot compute result!"
        assert (
            self.result is not _unset
        ), "ChangeWatcher.__exit__ was not called, cannot compute result!"
        return f"{self.orig!r} changed to {self.result!r}"

    __repr__ = __str__



import contextlib
from typing import Any, Callable, Iterator, Literal, Type, TypeVar, cast, overload

from johen.config import ParametrizeConfig, compile_matchers, get_base_config
from johen.generators.annotations import AnnotationMatcher, AnnotationProcessingContext
from johen.random import gen

global_config: ParametrizeConfig = get_base_config()

_A = TypeVar("_A")


@overload
def generate(
    obj: Type[_A],
    generate_defaults: bool | Literal["holes"] | None = None,
    matchers: list[AnnotationMatcher] | None = None,
    seed: int | None = None,
    count: int | None = None,
    globals: dict[str, Any] | None = None,
) -> Iterator[_A]:
    ...


@overload
def generate(
    obj: Callable[..., _A],
    generate_defaults: bool | Literal["holes"] | None = None,
    matchers: list[AnnotationMatcher] | None = None,
    seed: int | None = None,
    count: int | None = None,
    globals: dict[str, Any] | None = None,
) -> Iterator[_A]:
    ...


@overload
def generate(
    obj: Any,
    generate_defaults: bool | Literal["holes"] | None = None,
    matchers: list[AnnotationMatcher] | None = None,
    seed: int | None = None,
    count: int | None = None,
    globals: dict[str, Any] | None = None,
) -> Iterator:
    ...


def generate(
    obj: Any,
    generate_defaults: bool | Literal["holes"] | None = None,
    matchers: list[AnnotationMatcher] | None = None,
    seed: int | None = None,
    count: int | None = None,
    globals: dict[str, Any] | None = None,
) -> Iterator:
    context = AnnotationProcessingContext.from_source(obj)

    if generate_defaults is not None:
        context.generate_defaults = generate_defaults
    else:
        context.generate_defaults = global_config["generate_defaults"]

    if matchers is not None:
        context.matchers = [*matchers, *compile_matchers(global_config)]
    else:
        context.matchers = compile_matchers(global_config)

    if globals is not None:
        context.globals = {**global_config["globals"], **globals}
    else:
        context.globals = global_config["globals"]

    if seed is not None:
        gen.restart_at(seed=seed)

    if count is not None:
        result = [i for i, _ in zip(context.generate(), range(count))]
        assert len(result) == count, f"Could not generate {count} values for {obj}"
        return iter(result)

    gen.remaining_iterations = global_config["max_iterations"]
    return context.generate()


@contextlib.contextmanager
def replace_global_config(new_config: ParametrizeConfig):
    orig = global_config.copy()
    cast(dict, global_config).clear()
    global_config.update(new_config)
    try:
        yield
    finally:
        cast(dict, global_config).clear()
        global_config.update(orig)



import dataclasses
import enum
import inspect
import itertools
import types
import typing
from typing import Any, Iterator, get_type_hints

from johen.examples import Examples
from johen.exc import GenerationError
from johen.generators.annotations import AnnotationProcessingContext
from johen.generators.specialized import SimpleSymbol, ints
from johen.random import gen

__all__ = [
    "generate_dicts",
    "generate_enums",
    "generate_lists_sets_frozen_sets",
    "generate_tuples",
    "generate_unions",
    "generate_annotated",
    "generate_literals",
    "generate_dicts_for_annotations",
    "generate_dataclass_instances",
    "generate_unexpected_annotation",
    "generate_dicts_for_dataclass_model",
    "generate_call_args_for_argspec",
    "generate_results_from_call",
    "generate_dicts_from_typeddict",
    "generate_named_tuples",
]


def generate_dicts_for_annotations(
    annotations: typing.Mapping[str, int],
    context: "AnnotationProcessingContext",
    optional_keys: list[str],
) -> Iterator[dict[str, Any]]:
    generators: dict[str, Iterator[Any]] = {
        k: context.step(v, k)
        for k, v in annotations.items()
        if context.generate_defaults or k not in optional_keys
    }

    if not generators:
        return ({} for _ in gen)

    for k, v in generators.items():
        assert hasattr(v, "__next__")

    sampled_keys: Iterator[list[str]]
    if context.generate_defaults == "holes":
        sampled_keys = (r.sample(optional_keys, r.randint(0, len(optional_keys))) for r in gen)
    elif context.generate_defaults:
        sampled_keys = itertools.repeat(optional_keys)
    else:
        sampled_keys = itertools.repeat([])

    return (
        dict(
            (k, v)
            for k, v in zip(generators.keys(), values)
            if k not in optional_keys or k in included_keys
        )
        for values, included_keys in zip(zip(*generators.values()), sampled_keys)
    )


def generate_dicts_for_dataclass_model(
    context: "AnnotationProcessingContext",
) -> Iterator[dict[str, Any]]:
    hints = get_type_hints(context.source, include_extras=True)
    fields = {f.name: f for f in dataclasses.fields(context.source)}
    return generate_dicts_for_annotations(
        {k: hints.get(k, Any) for k, field in fields.items()},
        context,
        optional_keys=[k for k, field in fields.items() if _dataclass_has_default(field)],
    )


class FullArgSpec(typing.NamedTuple):
    args: list[SimpleSymbol]
    varargs: SimpleSymbol | None
    varkw: SimpleSymbol | None
    defaults: tuple[typing.Annotated[Any, Examples(ints)], ...] | None
    kwonlyargs: list[SimpleSymbol]
    kwonlydefaults: dict[SimpleSymbol, typing.Annotated[Any, Examples(ints)]] | None
    annotations: dict[str, typing.Annotated[Any, Examples((int,))]]


def generate_call_args_for_argspec(
    source: FullArgSpec, context: "AnnotationProcessingContext"
) -> typing.Iterator[tuple[tuple[Any, ...], dict[str, Any]]]:
    num_arg_defaults = len(source.defaults) if source.defaults is not None else 0

    required_args: list[Iterator[Any]] = []
    non_required_args: list[Iterator[Any]] = []
    for i, argname in enumerate(source.args):
        arg_iter = context.step(source.annotations.get(argname, Any), argname)
        if i < (len(source.args) - num_arg_defaults):
            required_args.append(arg_iter)
        elif context.generate_defaults:
            non_required_args.append(arg_iter)

    arg_generator: Iterator[tuple[Any, ...]]
    if required_args:
        arg_generator = zip(*required_args)
    else:
        arg_generator = itertools.repeat(tuple())

    if context.generate_defaults and non_required_args:
        arg_generator = (
            (
                *required,
                *(
                    non_required
                    if context.generate_defaults is True
                    else non_required[: r.randint(0, len(non_required))]
                ),
            )
            for required, non_required, r in zip(arg_generator, zip(*non_required_args), gen)
        )

    if source.varargs and context.generate_defaults:
        star_args_generator = context.step(
            tuple, source.varargs, (source.annotations.get(source.varargs, Any), ...)
        )
        arg_generator = (
            (*argset, *star_argset)
            for argset, star_argset in zip(arg_generator, star_args_generator)
        )

    kwargs_generator: Iterator[dict[str, Any]] = generate_dicts_for_annotations(
        {k: source.annotations.get(k, Any) for k in source.kwonlyargs},
        context,
        optional_keys=[
            k for k in source.kwonlyargs if source.kwonlydefaults and k in source.kwonlydefaults
        ],
    )

    if source.varkw and context.generate_defaults:
        kw_type = source.annotations.get(source.varkw, Any)
        origin = typing.get_origin(kw_type)
        args = typing.get_args(kw_type)

        varkw_generator: Iterator[dict[str, Any]]
        if origin is None or origin is not typing.Unpack:
            varkw_generator = context.step(dict, source.varkw, (str, kw_type))
        else:
            varkw_generator = context.step((*args, Any)[0], source.varkw)

        if varkw_generator:
            kwargs_generator = (
                dict(**args, **{k: v for k, v in kwargs.items() if k not in args})
                for args, kwargs in zip(kwargs_generator, varkw_generator)
            )

    return zip(arg_generator, kwargs_generator)


def generate_enums(context: "AnnotationProcessingContext") -> Iterator[Any] | None:
    if context.concretely_implements(enum.Enum) or context.concretely_implements(enum.IntEnum):
        return gen.one_of(context.source)
    return None


def generate_literals(context: "AnnotationProcessingContext") -> Iterator[Any] | None:
    if context.origin is typing.Literal:
        return gen.one_of(context.args)
    return None


def generate_dataclass_instances(context: AnnotationProcessingContext) -> Iterator[Any] | None:
    if dataclasses.is_dataclass(context.source):
        return (context.source(**d) for d in generate_dicts_for_dataclass_model(context))
    return None


def generate_dicts_from_typeddict(context: AnnotationProcessingContext) -> Iterator[Any] | None:
    is_match = typing.is_typeddict(context.source)
    if not is_match:
        try:
            import typing_extensions

            is_match = typing_extensions.is_typeddict(context.source)
        except ImportError:
            pass

    if is_match:
        optional: list[str] = sorted(getattr(context.source, "__optional_keys__", frozenset()))
        hints = get_type_hints(context.source, include_extras=True)
        return generate_dicts_for_annotations(
            {k: v for k, v in hints.items()}, context, optional_keys=list(optional)
        )

    return None


def generate_dicts(context: AnnotationProcessingContext) -> Iterator[dict[Any, Any]] | None:
    if context.concretely_implemented_by(dict):
        key, value, *_ = (*context.args, str, str)
        key_generator = context.step(key, "[Key]")
        value_generator = context.step(value, "[Value]")
        return context.wrap_with_debug_context(
            dict((k, v) for k, v, _ in zip(key_generator, value_generator, range(length)))
            for length in (r.randint(0, 5 - context.recursive_depth) for r in gen)
        )

    return None


def generate_results_from_call(context: AnnotationProcessingContext) -> Iterator[Any] | None:
    if inspect.isfunction(context.source):
        argspec = inspect.getfullargspec(context.source)
        return (
            context.source(*a, **k)
            for a, k in generate_call_args_for_argspec(FullArgSpec(*argspec), context)
        )

    return None


def generate_named_tuples(context: AnnotationProcessingContext) -> Iterator[Any] | None:
    if context.concretely_implements(tuple) and hasattr(context.source, "_field_defaults"):
        keys: tuple[str, ...] = context.source._fields  # noqa
        defaults: dict[str, Any] = context.source._field_defaults
        hints = get_type_hints(context.source, include_extras=True)
        dicts = generate_dicts_for_annotations(
            {k: hints.get(k, Any) for k in keys}, context, optional_keys=list(defaults.keys())
        )
        return (context.source(**d) for d in dicts)
    return None


def generate_tuples(context: AnnotationProcessingContext) -> Iterator[Any] | None:
    if context.concretely_implemented_by(tuple):
        has_ellipsis = not context.args or Ellipsis in context.args
        specified_parts = tuple(a for a in context.args if a is not Ellipsis)

        generators = [context.step(arg, str(f"[{i}]")) for i, arg in enumerate(specified_parts)]

        if generators:
            specified_generator = zip(*generators)
        else:
            specified_generator = itertools.repeat(())  # type: ignore

        if has_ellipsis:
            extension_type = [Any, *specified_parts][-1]
            unspecified_generator = context.step(list, "...", (extension_type,))
            return (
                (*specified, *unspecified)
                for specified, unspecified in zip(specified_generator, unspecified_generator)
            )
        else:
            return specified_generator
    return None


def generate_unions(context: AnnotationProcessingContext) -> Iterator[Any] | None:
    if context.origin in (typing.Union, types.UnionType) and context.args:
        return gen.one_of(*(context.step(arg, f"|") for arg in context.args))
    return None


def generate_lists_sets_frozen_sets(context: AnnotationProcessingContext) -> Iterator[Any] | None:
    for constructor in (list, set, frozenset):
        if context.concretely_implemented_by(constructor):
            arg = next(iter(context.args), Any)
            generator = context.step(arg)
            return (
                constructor(
                    [i for i, _ in zip(generator, range(r.randint(0, 5 - context.recursive_depth)))]
                )
                for r in gen
            )
    return None


def generate_unexpected_annotation(
    context: AnnotationProcessingContext,
) -> typing.Iterator[Any] | None:
    # Assume that the first argument is the actual type to generate
    if context.origin is not None and context.args is not None:
        annotated_inner = next(iter(context.args), Any)
        return context.step(annotated_inner)
    return None


def generate_annotated(context: AnnotationProcessingContext) -> typing.Iterator[Any] | None:
    if context.origin is typing.Annotated:
        annotated_inner = [*context.args, Any][0]
        examples = (v for ex in context.args[1:] if isinstance(ex, Examples) for v in ex)
        if examples:
            return gen.one_of(*examples)

        return context.step(annotated_inner)
    return None


def _dataclass_has_default(field: dataclasses.Field) -> bool:
    return (
        field.default is not dataclasses.MISSING or field.default_factory is not dataclasses.MISSING
    )


def generate_forward_refs(context: AnnotationProcessingContext) -> typing.Iterator | None:
    match = isinstance(context.source, typing.ForwardRef)
    if not match:
        try:
            import typing_extensions

            match = isinstance(context.source, typing_extensions.ForwardRef)
        except ImportError:
            pass
    if match:
        ref = typing.cast(typing.ForwardRef, context.source)
        if ref.__forward_arg__ not in context.globals:
            raise GenerationError(f"Could not resolve forward ref {ref.__forward_arg__}")

        def generate_for_forward_ref():
            t = ref._evaluate(context.globals, locals(), frozenset())
            for i in context.step(t, ref.__forward_arg__, recursive=True):
                yield i

        return generate_for_forward_ref()
    return None



from setuptools import find_packages, setup


def get_requirements(file: str):
    with open(file) as fp:
        return [x.strip() for x in fp.read().split("\n") if not x.startswith("#")]


setup(
    name="johen",
    version="0.1.2",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    install_requires=[],
    setup_requires=[],
    tests_require=get_requirements("test-requirements.txt"),
    classifiers=[
        "Intended Audience :: Developers",
        "Framework :: Pytest",
        "Topic :: Software Development :: Testing",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development",
    ],
    entry_points={"pytest11": ["johen = johen.pytest"]},
)



from johen.change_watcher import change_watcher
from johen.exc import GenerationError
from johen.globals import generate, global_config, replace_global_config
from johen.random import gen

__all__ = [
    "gen",
    "global_config",
    "generate",
    "replace_global_config",
    "change_watcher",
    "GenerationError",
]



import dataclasses
import random
import typing
from typing import Iterator

_A = typing.TypeVar("_A")


__all__ = ["gen"]


@dataclasses.dataclass
class _RandomGenerator:
    """
    A global, nearly-never ending generator for producing Random instances to other random source generators.
    Allows its generation to be seeded, ensuring that consumers of this iterator will behave deterministically
    given that they draw all their source of random variables from this iterator.

    Use the `gen` variable directly to construct such sources of generative data:
    my_integers = (r.randint(0, 100) for r in gen)
    """

    last_seed: int = 0
    r: random.Random = dataclasses.field(default_factory=lambda: random.Random(0))
    # A control valve that breaks the halting problem in the case of a bug, or at least highly costly generative step.
    remaining_iterations: int = (
        -1
    )  # Reset this before generating each parameter at the top of a process.

    def restart_at(self, seed: int) -> typing.Self:
        self.last_seed = seed
        self.r = random.Random(seed)
        return self

    def wrap_deterministically(
        self, iter: Iterator[_A], seed: int, max_iterations: int
    ) -> Iterator[_A]:
        """
        Enforces deterministic seed resolution, and a max_iterations per top level yield.
        """

        def wrapped():
            self.restart_at(seed)
            self.remaining_iterations = max_iterations
            for rv in iter:
                yield rv
                self.restart_at_next_seed()
                self.remaining_iterations = max_iterations

        return wrapped()

    def restart_at_next_seed(self) -> typing.Self:
        self.restart_at(self.last_seed)
        next_seed = self.r.getrandbits(64)
        self.restart_at(next_seed)
        return self

    def __next__(self) -> "random.Random":
        if self.remaining_iterations <= 0:
            raise StopIteration("Could not find generation")
        self.remaining_iterations -= 1
        return self.r

    def __iter__(self) -> "Iterator[random.Random]":
        return self

    @staticmethod
    def one_of(*options: typing.Iterable[_A]) -> Iterator[_A]:
        composed_options: list[typing.Iterable[_A]] = [
            _RandomGenerator._normalize(i) for i in options
        ]
        return (next(iter(r.choice(composed_options))) for r in gen)

    @staticmethod
    def _normalize(i: typing.Iterable[_A]) -> typing.Iterator[_A]:
        if not hasattr(i, "__next__"):
            i = list(i)
            return (r.choice(i) for r in gen)
        return iter(i)


gen = _RandomGenerator()
